/************************************************************************
**
**  Copyright (C) 2005  Christian Kranz
**
**  Siemens S65 Display Control
**
**  modded by Juergen Eckert
**
**  re-moddes by Marc Seiffert 
**
**  Circle-module given by Harry Sk. developed by unknown
**
************************************************************************/

#include "s65.h"
#include <avr/io.h> 
#include <util/delay.h>
#include <avr/pgmspace.h>

// GLOBALE VARIABLEN, benutzervernderbar
uint8_t 	*FontPointer; 	//FontPointer auf Font
uint8_t 	rot;			//Ausrichtung der Schrift 0=0, 1=90
uint16_t 	TextFarbe;		//Textfarbe, 5-6-5 RGB
uint16_t 	HgFarbe;		//Textfarbe, 5-6-5 RGB
uint8_t 	TextTyp;		//Texttyp 0=Proportional , 1=Fixed	
uint8_t		FNull;			// Fhrende Nullen bei Zahlen anzeigen	0=Aus , 1=an
uint8_t		NonTransparenz; // Transparenz der Schriftarten. Wenn aktiv ist Weiss =  Transparent   0=An , 1=Aus
uint8_t		BmpTransparenz;	// Transparenz bei Bitmaps --> Achtung, verschlechtert die Zeichenperformance um den Faktor 7 !
uint8_t		BmpTransFarbe;	// Farbe, die transparent dargestellt wird
// GLOBAELE VARIABLEN, zur internen Benutzung
uint8_t		TextBreite;		// Maximale Textbreite der aktuellen Font
char		za[20];			// Zahlenzwischenspeicher
// GLOBALE VARIABLEN ende

// Initialisierungssequenzen
prog_uint8_t INIT1[24] = {
0xFD, 0xFD, 0xFD, 0xFD, 0xEF, 0x00, 0xEE, 0x04, 
0x1B, 0x04, 0xFE, 0xFE, 0xFE, 0xFE, 0xEF, 0x90, 
0x4A, 0x04, 0x7F, 0x3F, 0xEE, 0x04, 0x43, 0x06,};
prog_uint8_t INIT2[40] = {
0xEF, 0x90, 0x09, 0x83, 0x08, 0x00, 0x0B, 0xAF, 
0x0A, 0x00, 0x05, 0x00, 0x06, 0x00, 0x07, 0x00, 
0xEF, 0x00, 0xEE, 0x0C, 0xEF, 0x90, 0x00, 0x80, 
0xEF, 0xB0, 0x49, 0x02, 0xEF, 0x00, 0x7F, 0x01, 
0xE1, 0x81, 0xE2, 0x02, 0xE2, 0x76, 0xE1, 0x83,};
prog_uint8_t INIT3[2] = {0x80, 0x01};
// Initialisierungssequenzen ende



void	S65_HgBelInit(uint8_t Helligkeit){
	PORTB &= ~_BV(PB7);  // clear port before enable
	DDRB |= _BV(PB7);  // will be used for OC2, must be output
	TCCR2 = _BV(WGM21) | _BV(WGM20) | _BV(COM21) | _BV(CS20);
	TCNT2=0x00;	OCR2=Helligkeit;
}

void	S65_HgBelSet(uint8_t Helligkeit){OCR2=Helligkeit;}


void LCD_WrCmd(unsigned char data){
	CLR(LCD_CS);					//select Display
	SET(LCD_RS);					//set RS line
	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
	SET(LCD_CS);
}

void LCD_WrData(unsigned char data){
	CLR(LCD_CS);					//select Display
	CLR(LCD_RS);					//clear RS line
	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
	SET(LCD_CS);
}

void mswait(unsigned int ms){
	unsigned int i;
	for(i=0; i<ms; i++)
		_delay_ms(1);
}

void wrtab(unsigned char* data, unsigned char size){
	unsigned char i;
	for(i=0; i<size; i++)
	{
		LCD_WrCmd(pgm_read_word_near(&data[i]));
	}
}

void S65_Init(void){
	OUTPUT(LCD_RESET);				//set reset as output
	CLR(LCD_RESET);					//Display reset
	INPUT(LCD_MISO);				//MISO is always input in SPI
	SET(LCD_MISO);					//enable pull up of MISO to avoid floating input
	OUTPUT(LCD_MOSI);				//MOSI must be set as output for SPI
	OUTPUT(LCD_CS);					//SS must be output for Master mode to work
	SET(LCD_CS);					//deselect display 
	OUTPUT(LCD_SCK);				//set SCK as output
	CLR(LCD_SCK);					//set SCK lo
	OUTPUT(LCD_RS);					//RS signal is output
	SET(LCD_RS);					//start with RS=high
		SPCR = (1<<MSTR) | (1<<SPE);//Master Mode,  Enable SPI, fCPU/4 clock
		SPSR |= (1<<SPI2X);			//double speed bit
	mswait(1);
		SET(LCD_RESET);				//clear status
		wrtab(INIT1, 24);
	mswait(7);
		wrtab(INIT2, 40);
	mswait(50);
		wrtab(INIT3, 2);
}

void S65_wrcmd16(uint16_t dat){LCD_WrCmd((dat>>8));LCD_WrCmd(dat);}

void S65_wrdat16(uint16_t dat){LCD_WrData(dat>>8);LCD_WrData(dat);}

void S65_allocatearea(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2){
	uint8_t x,y;
	if(x1 < x2){ x = x2; x2 = x1; x1 = x;} // x2 muss grer sein als x1
	if(y2 < y1){ y = y1; y1 = y2; y2 = y;} // x2 muss grer sein als x1
	S65_wrcmd16(0xEF80);		S65_wrcmd16(0x1805);
	S65_wrcmd16(0x1200 + x1);	S65_wrcmd16(0x1500 + x2);
	S65_wrcmd16(0x1300 + y1);	S65_wrcmd16(0x1600 + y2);
}

void S65_Fillscreen(uint16_t color)
{
  	S65_wrcmd16(0xEF90);  	S65_wrcmd16(0x0500);
  	S65_wrcmd16(0x0600);  	S65_wrcmd16(0x0700);
   	for (uint16_t i=0; i<S65_DISP_W*S65_DISP_H; i++){S65_wrdat16(color);}
}

void S65_DrawLine(uint8_t py1,uint8_t px1,uint8_t py2,uint8_t px2,uint16_t c){
  int16_t x, y, x1, x2, y1, y2, m, ofs;
  x1 = (uint16_t)px1; x2 = (uint16_t)px2;                 	// Daten bertragen
  y1 = (uint16_t)py1; y2 = (uint16_t)py2;                 	// Daten bertragen
	if(x2 < x1){                                         	// x2 muss grer sein als x1
		x = x1; x1 = x2; x2 = x;
		y = y1; y1 = y2; y2 = y;
	};

	if (y2 == y1){                                        	// Horizontale linie
		for(uint8_t counter = x1;counter <= x2;counter++){    	// Von X1 nach X2
			S65_Pixel((int8_t)counter, (int8_t)y1, c);  	// Punkt setzen
			};
		}
	else{
		if(x2 == x1){                                     	// Vertikale linie
			if(y2 < y1){                                  	// Y2 muss grer sein als Y1
				x = x1; x1 = x2; x2 = x;
				y = y1; y1 = y2; y2 = y;
			};
			for(uint8_t counter = y1; counter<= y2;counter++){	// Von Y1 nach Y2
				S65_Pixel((uint8_t)x1,(uint8_t)counter,c);	// Punkt setzen
			};
		}
		else{
			m = (127*(y2 - y1)) / (x2 - x1);              	// Steigung ausrechnen
			if((m < 127) && (m > - 127)){                		// <= 45 Winkel
			m = (500*(y2 - y1)) / (x2 - x1);                 	// Steigung ausrechnen
				if( m >= 0){
					ofs = 249;}
				else{
					ofs = - 249;};  			// offset setzen
				for(uint8_t counter=0;counter<=(x2-x1);counter++){// von 0 nach Lnge
					y = (counter*m + ofs) / 500;                  	// Y-Wert berechnen
					S65_Pixel((uint8_t)counter+(uint8_t)x1,(uint8_t)y+y1,c);	// Punkt setzen
				};
			}
			else{
				if((m >= 127) || (m <= - 127)){                   	// > 45 Winkel
					if(y2 < y1){                                  	// Y2 muss grer sein als Y1
						x = x1; x1 = x2; x2 = x;
						y = y1; y1 = y2; y2 = y;
					};
					m = (255*(y2 - y1)) / (x2 - x1);                	// Steigung berechnen
					for(uint8_t counter=0;counter<=(y2-y1);counter++){	// von 0 nach hhe
						x = ((counter*255) + 127) / m;              	// X-Wert berechnen
						S65_Pixel((uint8_t)(x+x1),(uint8_t)counter+(uint8_t)y1,c);	// Punkt setzen
					}
				}
			}
		}
	}
}

void S65_Pixel(uint8_t y, uint8_t x, uint16_t color){
	S65_allocatearea(x,y,x,y);
	S65_wrdat16(color);
}

void S65_DrawRect(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2,uint16_t c){
  uint8_t x,y;
  if(x2 < x1){                                          // x2 muss grer sein als x1
    x = x1; x1 = x2; x2 = x;
    y = y1; y1 = y2; y2 = y;
  }
  for(uint8_t counter = x1; counter <= x2; counter++){      // X-Seiten abf8x8hren
    S65_Pixel(counter, y1, c);                        // obere H-Linie
    S65_Pixel(counter, y2, c);                        // untere H-Linie
  }
  for(uint8_t counter = y1; counter <= y2; counter++){                               // Y-Seiten abf8x8hren
    S65_Pixel(x1, counter, c);                        // linke V-Linie
    S65_Pixel(x2, counter, c);                        // rechte V-Linie
	}
}

void S65_DrawFilledRect(uint16_t x1,uint16_t y1,uint16_t x2,uint16_t y2,uint16_t color){
	S65_allocatearea(x1,y1,x2,y2);
		for(uint16_t counter = 0; counter < ((1+x2-x1)*(1+y2+y1)); counter++){
	S65_wrdat16(color);
	}
}

void S65_Circle(uint8_t x0, uint8_t y0, uint8_t radius, uint16_t color){
   int f = 1 - radius;
   int ddF_x = 0;
   int ddF_y = -2 * radius;
   int x = 0;
   int y = radius;
	   S65_Pixel(x0, y0 + radius,color);
	   S65_Pixel(x0, y0 - radius,color);
	   S65_Pixel(x0 + radius, y0,color);
	   S65_Pixel(x0 - radius, y0,color);
   while(x < y)
   {
     if(f >= 0)
     {
       y--;
       ddF_y += 2;
       f += ddF_y;
     }
     x++;
     ddF_x += 2;
     f += ddF_x + 1;
	     S65_Pixel(x0 + x, y0 + y,color);	     S65_Pixel(x0 - x, y0 + y,color);
	     S65_Pixel(x0 + x, y0 - y,color);	     S65_Pixel(x0 - x, y0 - y,color);
	     S65_Pixel(x0 + y, y0 + x,color);	     S65_Pixel(x0 - y, y0 + x,color);
	     S65_Pixel(x0 + y, y0 - x,color);	     S65_Pixel(x0 - y, y0 - x,color);
   }
}

void 	S65_SetFont(uint8_t *pointer){FontPointer = pointer;TextBreite = pgm_read_byte_near(&FontPointer[2]);}
void 	S65_SetRot(uint8_t r){rot = r;};
void 	S65_SetTextColor(uint16_t color){TextFarbe = color;};
void 	S65_SetTextTyp(uint8_t typ){TextTyp = typ;}
void 	S65_SetFNull(uint8_t fn){FNull = fn;};
void	S65_SetNonTransparenz(uint8_t nt){NonTransparenz = nt;};
void	S65_SetBmpTransparenz(uint8_t bt){BmpTransparenz = bt;};
void 	S65_SetBmpTransFarbe(uint16_t color){BmpTransFarbe = color;};

uint8_t S65_PutChar(uint8_t x, uint8_t y, uint8_t c) // Ein Zeichen ausgeben, vorher muss eine Schriftart geladen werden
{
	uint8_t 	hoehe 			=	pgm_read_byte_near(&FontPointer[3]);
	uint8_t 	firstchar 		=	pgm_read_byte_near(&FontPointer[5]);
	uint8_t 	lastchar 		=	pgm_read_byte_near(&FontPointer[6]);
	uint8_t 	zeichenbreite 	= 	pgm_read_byte_near(&FontPointer[7+c-firstchar]);
	uint8_t 	byte 			= 	0;
	uint8_t 	bitoffset 		= 	0;
	uint8_t 	maske 			= 	0;
	uint8_t 	bcounter 		= 	0;
	uint16_t 	bitsbischar 	= 	0;
	uint16_t 	bytesbischar	= 	0;
	uint16_t 	counter 		= 	0;

	for(uint8_t Ccounter = 0; Ccounter < c-firstchar; Ccounter++)
		{
		bitsbischar += (pgm_read_byte_near(&FontPointer[Ccounter+7]));
		};

	bitsbischar *= hoehe;
	bitsbischar += ((lastchar-firstchar)+7)*8;
	bytesbischar = bitsbischar/8;
	bitoffset = bitsbischar % 8;
	maske = bitoffset % 8;

	for(uint8_t xc = 0; xc < zeichenbreite; xc++) // Eigentlcihe Ausgaberoutine
		{
		for(uint8_t yc = hoehe; yc > 0; yc--)
			{
			if(maske > 7){maske = 0;bcounter+=1;};
			byte = pgm_read_byte_near(&FontPointer[bytesbischar + bcounter]+1);
			
			if(bit_is_set(byte,maske))
				{
				if(rot)
					{
					S65_Pixel((hoehe-yc)+y,xc+x, TextFarbe);
					}
				else
					{
					S65_Pixel(xc+x,yc+y, TextFarbe);
					}
				}
			else
				{
				if(NonTransparenz)
				{
					if(rot)
						{
						S65_Pixel((hoehe-yc)+y,xc+x, HgFarbe);
						}
					else
						{
						S65_Pixel(xc+x,yc+y, HgFarbe);
						}
					}
				}

			maske++;
			counter++;
			}
		}
	return(zeichenbreite);
}

uint8_t	S65_PutStr(uint8_t x, uint8_t y, char *c) // Einen Nulterimierten String ausgeben
{	
	char zeichen = 0;
	uint8_t zeichenbreite = 0;
	uint8_t xo = 0;
	uint8_t counter = 0;
	while(c[counter] != 0)
		{
		zeichen = c[counter];
		zeichenbreite = S65_PutChar(x+xo,y,zeichen);
		if(TextTyp == 0)
			{
			xo += zeichenbreite+2;
			}
		else
			{
			xo+=TextBreite;
			}
		counter++;
		if(counter == 255){return(255);} // 255 ist "Fehleroode" fr unterminierten String
		}
	return(xo);
	}

void S65_PutByte(uint8_t x, uint8_t y, uint8_t z) // Routine zum ausgeben von Zahlen
	{
	za[3] = 0;
	za[2] = (z % 10)+48;z = z/10;
	za[1] = (z % 10)+48;z = z/10;
	za[0] = (z)+48;
	uint8_t counter = 0;
	if(!FNull)										// Fhrende Nullen entfernen / lassen
		{
		while(za[counter] == 48){counter++;}
		}
	S65_PutStr(x, y, &za[counter]);
	}

void S65_PutInt(uint8_t x, uint8_t y, uint16_t z)// Routine zum ausgeben von Zahlen
	{ 
	za[6] = 0;
	za[5] = (z % 10)+48;z = z/10;
	za[4] = (z % 10)+48;z = z/10;
	za[3] = (z % 10)+48;z = z/10;
	za[2] = (z % 10)+48;z = z/10;
	za[1] = (z % 10)+48;z = z/10;
	za[0] = (z)+48;
	uint8_t counter = 0;
	if(!FNull)										// Fhrende Nullen entfernen / lassen
		{
		while(za[counter] == 48){counter++;}
		}
	S65_PutStr(x, y, &za[counter]);
	}

void S65_PutLong(uint8_t x, uint8_t y, uint32_t z) // Routine zum ausgeben von Zahlen
	{
	za[13] =0;						za[12] =(z % 10)+48;z = z/10;	
	za[11] =(z % 10)+48;z = z/10;	za[10] =(z % 10)+48;z = z/10;	
	za[9] = (z % 10)+48;z = z/10;	za[8] = (z % 10)+48;z = z/10;	
	za[7] = (z % 10)+48;z = z/10;	za[6] = (z % 10)+48;z = z/10;	
	za[5] = (z % 10)+48;z = z/10;	za[4] = (z % 10)+48;z = z/10;	
	za[3] = (z % 10)+48;z = z/10;	za[2] = (z % 10)+48;z = z/10;	
	za[1] = (z % 10)+48;z = z/10;	za[0] = (z)+48;
	uint8_t counter = 0;
	if(!FNull)										// Fhrende Nullen entfernen / lassen
		{
		while(za[counter] == 48){counter++;}
		}
	S65_PutStr(x, y, &za[counter]);
	}

void	S65_PutRawBmp16(uint8_t x, uint8_t y, uint8_t *pic, uint8_t w, uint8_t h)
	{
	uint8_t dimx = w-1; 
	uint8_t dimy = h-1;
	S65_allocatearea(x,y,x+dimx, y+dimy);
	for(uint16_t counter = 0; counter < ((dimx+1)*(dimy+1))*2; counter++)
		{
		LCD_WrData(pgm_read_byte_near(&pic[counter]));
		}
	}

void	S65_PutBmp(uint8_t x, uint8_t y, uint8_t *pic)
	{
	uint8_t dimx= pgm_read_byte_near(&pic[0])-1; 
	uint8_t dimy= pgm_read_byte_near(&pic[1])-1;
	uint8_t BpP	= pgm_read_byte_near(&pic[2]);

	if(BpP==16) // Fullclor-Bitmap, die einfachste
		{
		if(BmpTransparenz) // Bitmap mit Transparenzfarbe zeichnen (~ 1 Vollbild /s)
			{
			uint16_t counter = 0;
			uint16_t Farbe = 0;
			for(uint8_t y = dimy+1; y > 0 ; y--)
				{
				for(uint8_t x = 0; x <= dimx; x++)
					{
					Farbe = pgm_read_byte_near(&pic[counter+2]);	
					Farbe *= 256;
					Farbe += pgm_read_byte_near(&pic[counter+3]);
					if(BmpTransFarbe != Farbe){S65_Pixel(x,y-1,Farbe);}
					counter +=2;
					}
				}
			}
		else	// Bitmap ohne Transparenzfarbe zeichnen (~ 7 Vollbilder /s)
			{
			S65_allocatearea(y,x,y+dimy, x+dimx);
			for(uint16_t counter = 0; counter < ((dimx+1)*(dimy+1))*2; counter++)
				{
				LCD_WrData(pgm_read_byte_near(&pic[counter+2]));
				}
			}
		}
	else // Bitmap mit 2 bis 2^15 Farben
		{
		uint8_t		byte = 0;
		uint8_t 	bitpointer = 0;
		uint16_t 	offset = 0;
		uint16_t 	aktcolor = 0;
		uint16_t 	colortablesize = 0;
		uint16_t 	Pixel = (dimx+1) * (dimy+1);
		uint16_t 	counter = 0;
		uint16_t 	bytepointer = 0;
		uint16_t 	colortableindex = 0;

		aktcolor	= 	pgm_read_byte_near(&pic[4]); 
		aktcolor	*= 	256;
		aktcolor	+= 	pgm_read_byte_near(&pic[3]); 
		colortablesize	= 	pgm_read_byte_near(&pic[6]); 
		colortablesize	*= 	256;
		colortablesize	+= 	pgm_read_byte_near(&pic[5]);

		offset = colortablesize + 7;
		
		S65_allocatearea(y,x,y+dimy, x+dimx);

		while(counter < Pixel)
			{
			byte = pgm_read_byte_near(&pic[offset+bytepointer]);
			if(bit_is_set(byte, bitpointer))
				{
				colortableindex = 0;
				for(uint8_t ccounter = 0; ccounter < BpP; ccounter++)
					{
					colortableindex *=2;
					bitpointer +=1;
					if(bitpointer == 8){bitpointer = 0;bytepointer+=1;}
					byte = pgm_read_byte_near(&pic[offset+bytepointer]);
					if(bit_is_set(byte, bitpointer)){colortableindex += 1;}
					}
				aktcolor = pgm_read_byte_near(&pic[7+colortableindex]);
				}
			else
				{
					S65_wrdat16(aktcolor);
					counter++;
					bitpointer +=1;
					if(bitpointer == 8){bitpointer = 0;bytepointer+=1;}
				}
			}
		}
	}













